"""unxt: Quantities in JAX.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

from jaxtyping import install_import_hook

from .setup_package import RUNTIME_TYPECHECKER

with install_import_hook("unxt", RUNTIME_TYPECHECKER):
    from . import quantity, unitsystems
    from ._src.dimensions.core import dimensions, dimensions_of
    from ._src.units.core import units, units_of
    from ._version import version as __version__
    from .quantity import *
    from .unitsystems import AbstractUnitSystem, unitsystem, unitsystem_of

from ._src import experimental  # noqa: F401

# isort: split
from . import _interop  # noqa: F401  # register interop

__all__ = [
    "__version__",
    # dimensions
    "dimensions",
    "dimensions_of",
    # units
    "units",
    "units_of",
    # units systems
    "unitsystems",  # module
    "AbstractUnitSystem",  # base class
    "unitsystem",  # convenience constructor
    "unitsystem_of",  # get the unit system
]
__all__ += quantity.__all__

# Clean up the namespace
del install_import_hook



"""unxt: Quantities in JAX.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""
# ruff:noqa: F403

from ._src.quantity.api import uconvert, ustrip
from ._src.quantity.base import AbstractQuantity, is_unit_convertible
from ._src.quantity.base_parametric import AbstractParametricQuantity
from ._src.quantity.core import Quantity
from ._src.quantity.fast import UncheckedQuantity

# isort: split
# Register dispatches and conversions
from ._src.quantity import compat, functional, register_dispatches, register_primitives

__all__: list[str] = [
    # Base
    "AbstractQuantity",
    # Fast
    "UncheckedQuantity",
    # Base Parametric
    "AbstractParametricQuantity",
    # Core
    "Quantity",
    # Functional
    "uconvert",
    "ustrip",
    "is_unit_convertible",
]


# Clean up namespace
del compat, functional, register_dispatches, register_primitives



"""Astropy units compatibility."""

__all__: list[str] = []

from typing import TypeAlias

import astropy.units as u
from plum import dispatch

AstropyUnits: TypeAlias = u.UnitBase

# ===================================================================
# Register dispatches


@dispatch
def units(obj: u.UnitBase | u.Unit, /) -> AstropyUnits:
    """Construct the units from an Astropy unit.

    Examples
    --------
    >>> import astropy.units as u
    >>> from unxt import units
    >>> units(u.km)
    Unit("km")

    """
    return u.Unit(obj)


@dispatch
def units(obj: u.Quantity, /) -> AstropyUnits:
    """Construct the units from an Astropy quantity.

    Examples
    --------
    >>> import astropy.units as u
    >>> from unxt import units
    >>> units(u.Quantity(2, "km"))
    Unit("2 km")

    """
    return u.Unit(obj)


# -------------------------------------------------------------------


@dispatch
def units_of(obj: u.UnitBase | u.Unit, /) -> AstropyUnits:
    """Return the units of an object.

    Examples
    --------
    >>> import astropy.units as u
    >>> from unxt import units_of

    >>> units_of(u.km)
    Unit("km")

    """
    return obj


@dispatch
def units_of(obj: u.Quantity, /) -> AstropyUnits:
    """Return the units of an Astropy quantity.

    Examples
    --------
    >>> import astropy.units as u
    >>> from unxt import units_of

    >>> units_of(u.Quantity(1, "km"))
    Unit("km")

    """
    return units_of(obj.unit)



"""Built-in unit systems."""

from __future__ import annotations

__all__ = ["DimensionlessUnitSystem", "LTMAUnitSystem"]

from dataclasses import dataclass
from typing import Annotated, TypeAlias, final
from typing_extensions import override

import astropy.units as u
from astropy.units import dimensionless_unscaled

from . import builtin_dimensions as ud  # noqa: TCH001
from .base import AbstractUnitSystem
from unxt._src.utils import SingletonMixin

Unit: TypeAlias = u.UnitBase


@final
@dataclass(frozen=True, slots=True)
class DimensionlessUnitSystem(SingletonMixin, AbstractUnitSystem):
    """A unit system with only dimensionless units.

    This is a singleton class.

    Examples
    --------
    >>> from unxt.unitsystems import DimensionlessUnitSystem
    >>> dims1 = DimensionlessUnitSystem()
    >>> dims2 = DimensionlessUnitSystem()
    >>> dims1 is dims2
    True

    """

    dimensionless: Annotated[Unit, ud.dimensionless] = dimensionless_unscaled

    def __repr__(self) -> str:
        return "DimensionlessUnitSystem()"

    @override
    def __str__(self) -> str:
        return self.__repr__()


@final
@dataclass(frozen=True, slots=True, repr=False)
class LTMAUnitSystem(AbstractUnitSystem):
    """Length, time, mass, angle unit system."""

    length: Annotated[Unit, ud.length]
    time: Annotated[Unit, ud.time]
    mass: Annotated[Unit, ud.mass]
    angle: Annotated[Unit, ud.angle]

    def __repr__(self) -> str:
        fs = ", ".join(map(str, self.base_units))
        return f"unitsystem({fs})"


@final
@dataclass(frozen=True, slots=True)
class SIUnitSystem(AbstractUnitSystem):
    """SI unit system + angles.

    Note: this is not part of the public API! Use the `si` instance (realization) from
    `unxt.unitsystems` instead.
    """

    # Base SI dimensions
    length: Annotated[Unit, ud.length]
    time: Annotated[Unit, ud.time]
    amount: Annotated[Unit, ud.amount]
    electric_current: Annotated[Unit, ud.current]
    temperature: Annotated[Unit, ud.temperature]
    luminous_intensity: Annotated[Unit, ud.luminous_intensity]
    mass: Annotated[Unit, ud.mass]
    # + angles
    angle: Annotated[Unit, ud.angle]


@final
@dataclass(frozen=True, slots=True)
class CGSUnitSystem(AbstractUnitSystem):
    """CGS unit system + angles.

    Note: this is not part of the public API! Use the `cgs` instance (realization) from
    `unxt.unitsystems` instead.
    """

    # Base CGS dimensions
    length: Annotated[Unit, ud.length]
    mass: Annotated[Unit, ud.mass]
    time: Annotated[Unit, ud.time]
    force: Annotated[Unit, ud.force]
    energy: Annotated[Unit, ud.energy]
    pressure: Annotated[Unit, ud.pressure]
    dynamic_viscosity: Annotated[Unit, ud.dynamic_viscosity]
    kinematic_viscosity: Annotated[Unit, ud.kinematic_viscosity]
    # + angles
    angle: Annotated[Unit, ud.angle]



"""Unit system utils."""

__all__: list[str] = []

import re
from typing import Any

from astropy.units import PhysicalType as Dimension
from plum import dispatch

from . import builtin_dimensions as bdims
from unxt._src.dimensions.core import dimensions_of
from unxt._src.typing_ext import Unit

# ------------------------------------


@dispatch.abstract
def get_dimension_name(pt: Any, /) -> str:
    """Get the dimension name from the object."""
    raise NotImplementedError  # pragma: no cover


@dispatch  # type: ignore[no-redef]
def get_dimension_name(pt: str, /) -> str:
    """Return the dimension name.

    Note that this does not check for the existence of that dimension.

    Examples
    --------
    >>> from unxt._src.units.system.utils import get_dimension_name

    >>> get_dimension_name("length")
    'length'

    >>> get_dimension_name("not real")
    'not_real'

    >>> try: get_dimension_name("*62")
    ... except ValueError as e: print(e)
    Input contains non-letter characters

    """
    # A regex search to match anything that's not a letter or a whitespace.
    if re.search(r"[^a-zA-Z_ ]", pt):
        msg = "Input contains non-letter characters"
        raise ValueError(msg)

    return pt.replace(" ", "_")


@dispatch  # type: ignore[no-redef]
def get_dimension_name(pt: Dimension, /) -> str:
    """Return the dimension name from a dimension.

    Examples
    --------
    >>> from unxt._src.units.system.utils import get_dimension_name
    >>> import astropy.units as u
    >>> get_dimension_name(u.get_physical_type("length"))
    'length'

    >>> get_dimension_name(u.get_physical_type("speed"))
    'speed'

    """
    # Note: this is not deterministic b/c ``_physical_type`` is a set
    #       that's why the `if` statement is needed.
    match pt:
        case bdims.speed:
            out = "speed"
        case _:
            out = get_dimension_name(next(iter(pt._physical_type)))  # noqa: SLF001
    return out


@dispatch  # type: ignore[no-redef]
def get_dimension_name(pt: Unit, /) -> str:
    """Return the dimension name from a unit.

    Examples
    --------
    >>> from unxt._src.units.system.utils import get_dimension_name
    >>> import astropy.units as u
    >>> get_dimension_name(u.km)
    'length'

    """
    return get_dimension_name(dimensions_of(pt))



"""Unitsystem compatibility."""

__all__: list[str] = []

from typing import Any, TypeAlias

import astropy.units as u
from astropy.coordinates import Angle as AstropyAngle, Distance as AstropyDistance
from astropy.units import Quantity as AstropyQuantity
from jaxtyping import Array
from packaging.version import Version
from plum import conversion_method, dispatch

import quaxed.numpy as jnp
from dataclassish import replace

from unxt import AbstractQuantity, Quantity, UncheckedQuantity
from unxt._interop.optional_deps import OptDeps

# ============================================================================
# AbstractQuantity

# -----------------
# Constructor


@AbstractQuantity.from_._f.register  # noqa: SLF001
def from_(
    cls: type[AbstractQuantity], value: AstropyQuantity, /, **kwargs: Any
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import astropy.units as u
    >>> from unxt import Quantity

    >>> Quantity.from_(u.Quantity(1, "m"))
    Quantity['length'](Array(1., dtype=float32), unit='m')

    """
    return cls(jnp.asarray(value.value, **kwargs), value.unit)


@AbstractQuantity.from_._f.register  # type: ignore[no-redef] # noqa: SLF001
def from_(
    cls: type[AbstractQuantity], value: AstropyQuantity, unit: Any, /, **kwargs: Any
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import astropy.units as u
    >>> from unxt import Quantity

    >>> Quantity.from_(u.Quantity(1, "m"), "cm")
    Quantity['length'](Array(100., dtype=float32), unit='cm')

    """
    return cls(jnp.asarray(value.to_value(unit), **kwargs), unit)


# -----------------
# Conversion Methods


@conversion_method(type_from=AbstractQuantity, type_to=AstropyQuantity)  # type: ignore[misc]
def convert_unxt_quantity_to_astropy_quantity(
    q: AbstractQuantity, /
) -> AstropyQuantity:
    """Convert a `unxt.AbstractQuantity` to a `astropy.units.Quantity`.

    Examples
    --------
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert
    >>> from unxt import Quantity

    >>> convert(Quantity(1.0, "cm"), AstropyQuantity)
    <Quantity 1. cm>

    """
    return AstropyQuantity(q.value, q.unit)


@conversion_method(type_from=AbstractQuantity, type_to=AstropyDistance)  # type: ignore[misc]
def convert_unxt_quantity_to_astropy_distance(
    q: AbstractQuantity, /
) -> AstropyDistance:
    """Convert a `unxt.AbstractQuantity` to a `astropy.coordinates.Distance`.

    Examples
    --------
    >>> from astropy.coordinates import Distance
    >>> from plum import convert
    >>> from unxt import Quantity

    >>> convert(Quantity(1.0, "cm"), Distance)
    <Distance 1. cm>

    """
    return AstropyDistance(q.value, q.unit)


@conversion_method(type_from=AbstractQuantity, type_to=AstropyAngle)  # type: ignore[misc]
def convert_unxt_quantity_to_astropy_angle(q: AbstractQuantity, /) -> AstropyAngle:
    """Convert a `unxt.AbstractQuantity` to a `astropy.coordinates.Angle`.

    Examples
    --------
    >>> from astropy.coordinates import Angle
    >>> from plum import convert
    >>> from unxt import Quantity

    >>> convert(Quantity(1.0, "radian"), Angle)
    <Angle 1. rad>

    """
    return AstropyAngle(q.value, q.unit)


# ============================================================================
# Quantity


@conversion_method(type_from=AstropyQuantity, type_to=Quantity)  # type: ignore[misc]
def convert_astropy_quantity_to_unxt_quantity(q: AstropyQuantity, /) -> Quantity:
    """Convert a `astropy.units.Quantity` to a `unxt.Quantity`.

    Examples
    --------
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert
    >>> from unxt import Quantity

    >>> convert(AstropyQuantity(1.0, "cm"), Quantity)
    Quantity['length'](Array(1., dtype=float32), unit='cm')

    """
    return Quantity(q.value, q.unit)


# ============================================================================
# UncheckedQuantity


@conversion_method(type_from=AstropyQuantity, type_to=UncheckedQuantity)  # type: ignore[misc]
def convert_astropy_quantity_to_unxt_uncheckedquantity(
    q: AstropyQuantity, /
) -> UncheckedQuantity:
    """Convert a `astropy.units.Quantity` to a `unxt.UncheckedQuantity`.

    Examples
    --------
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert
    >>> from unxt import UncheckedQuantity

    >>> convert(AstropyQuantity(1.0, "cm"), UncheckedQuantity)
    UncheckedQuantity(Array(1., dtype=float32), unit='cm')

    """
    return UncheckedQuantity(q.value, q.unit)


###############################################################################


AstropyUnit: TypeAlias = u.UnitBase | u.Unit | u.FunctionUnitBase | u.StructuredUnit


if Version("7.0") <= OptDeps.ASTROPY.version:

    def _apy7_unit_to(self: AstropyUnit, other: AstropyUnit, value: Array, /) -> Array:
        return self.to(other, value)

else:

    def _apy7_unit_to(self: AstropyUnit, other: AstropyUnit, value: Array, /) -> Array:
        """Convert the value to the other unit."""
        # return self.get_converter(Unit(other), equivalencies)(value)
        # First see if it is just a scaling.
        try:
            scale = self._to(other)
        except u.UnitsError:
            pass
        else:
            return scale * value

        # if that doesn't work, maybe we can do it with equivalencies?
        try:
            return self._apply_equivalencies(
                self, other, self._normalize_equivalencies([])
            )(value)
        except u.UnitsError as exc:
            # Last hope: maybe other knows how to do it?
            # We assume the equivalencies have the unit itself as first item.
            # TODO: maybe better for other to have a `_back_converter` method?
            if hasattr(other, "equivalencies"):
                for funit, tunit, _, b in other.equivalencies:
                    if other is funit:
                        try:
                            converter = self.get_converter(tunit, [])
                        except Exception:  # noqa: BLE001, S110  # pylint: disable=W0718
                            pass
                        else:
                            return b(converter(value))

            raise exc  # noqa: TRY201


@dispatch  # type: ignore[misc]
def uconvert(unit: AstropyUnit, x: AbstractQuantity, /) -> AbstractQuantity:
    """Convert the quantity to the specified units.

    Examples
    --------
    >>> import astropy.units as u
    >>> from unxt import Quantity, units

    >>> x = Quantity(1000, "m")
    >>> uconvert(units("km"), x)
    Quantity['length'](Array(1., dtype=float32, ...), unit='km')

    >>> x = Quantity([1, 2, 3], "Kelvin")
    >>> with u.add_enabled_equivalencies(u.temperature()):
    ...     y = x.to(u.deg_C)
    >>> y
    Quantity['temperature'](Array([-272.15, -271.15, -270.15], dtype=float32), unit='deg_C')

    """  # noqa: E501
    # Hot-path: if no unit conversion is necessary
    if x.unit == unit:
        return x

    # TODO: jaxpr units so we can understand them at trace time.
    # Hot-path: if in tracing mode
    # if isinstance(x.value, jax.core.Tracer) and not is_unit_convertible(x.unit, u):
    #     return x.value

    return replace(x, value=_apy7_unit_to(x.unit, unit, x.value), unit=unit)



# pylint: disable=import-error, no-member, unsubscriptable-object
#    b/c it doesn't understand dataclass fields

__all__ = ["AbstractParametricQuantity"]

from collections.abc import Sequence
from functools import partial
from typing import Any

import equinox as eqx
import jax
import jax.core
from astropy.units import PhysicalType as Dimensions, Unit
from jaxtyping import Array, ArrayLike, Shaped
from plum import dispatch, parametric, type_nonparametric, type_unparametrized

from dataclassish import field_items, fields

from .base import AbstractQuantity
from unxt._src.dimensions.core import dimensions, dimensions_of
from unxt._src.typing_ext import Unit as UnitTypes
from unxt._src.units.core import units


@parametric
class AbstractParametricQuantity(AbstractQuantity):
    """Arrays with associated units.

    This class is parametrized by the dimensions of the units.

    """

    value: Shaped[Array, "*shape"] = eqx.field(converter=jax.numpy.asarray)
    """The value of the `AbstractQuantity`."""

    unit: Unit = eqx.field(static=True, converter=units)
    """The unit associated with this value."""

    def __post_init__(self) -> None:
        """Check whether the arguments are valid."""
        self._type_parameter: Dimensions

    def __check_init__(self) -> None:
        """Check whether the arguments are valid."""
        expected_dimensions = self._type_parameter
        got_dimensions = dimensions_of(self.unit)
        if got_dimensions != expected_dimensions:
            msg = "Physical type mismatch."  # TODO: better error message
            raise ValueError(msg)

    # ---------------------------------------------------------------
    # Plum stuff

    @classmethod
    @dispatch
    def __init_type_parameter__(cls, dims: Dimensions, /) -> tuple[Dimensions]:
        """Check whether the type parameters are valid."""
        return (dims,)

    @classmethod  # type: ignore[no-redef]
    @dispatch
    def __init_type_parameter__(cls, dims: str, /) -> tuple[Dimensions]:
        """Check whether the type parameters are valid."""
        dims_: Dimensions
        try:
            dims_ = dimensions(dims)
        except ValueError:
            dims_ = dimensions_of(units(dims))
        return (dims_,)

    @classmethod  # type: ignore[no-redef]
    @dispatch
    def __init_type_parameter__(cls, unit: UnitTypes, /) -> tuple[Dimensions]:
        """Infer the type parameter from the arguments."""
        dims = dimensions_of(unit)
        if dims != "unknown":
            return (dims,)
        return (Dimensions(unit, unit.to_string(fraction=False)),)

    @classmethod
    def __infer_type_parameter__(
        cls, value: ArrayLike | Sequence[Any], unit: Any, **kwargs: Any
    ) -> tuple[Dimensions]:
        """Infer the type parameter from the arguments."""
        return (dimensions_of(units(unit)),)

    @classmethod
    @dispatch  # type: ignore[misc]
    def __le_type_parameter__(
        cls, left: tuple[Dimensions], right: tuple[Dimensions]
    ) -> bool:
        """Define an order on type parameters.

        That is, check whether `left <= right` or not.
        """
        (dim_left,) = left
        (dim_right,) = right
        return dim_left == dim_right

    # ---------------------------------------------------------------
    # misc

    def __getnewargs_ex__(self) -> tuple[tuple[Any, ...], dict[str, Any]]:
        """Return args, kwargs for ``__new__``.

        In protocols 2+, this is used to determine the values (args,
        and kwargs) passed to ``__new__``. We implement ``__getnewargs_ex__``
        instead of ``__getnewargs__`` because the latter does not support
        keyword-only arguments.

        Examples
        --------
        >>> import copy as pycopy
        >>> from unxt import Quantity
        >>> x = Quantity([1, 2, 3], "m")
        >>> pycopy.copy(x)
        Quantity['length'](Array([1, 2, 3], dtype=int32), unit='m')

        """
        return (), field_items(self)

    # TODO: see if pickling can be accomplished without reduce.
    # https://docs.python.org/3.12/library/pickle.html
    def __reduce__(
        self,
    ) -> tuple[partial["type[AbstractParametricQuantity]"], tuple[Any, ...]]:
        r"""Return the reduced value: a constructor and arguments.

        The ``__reduce__`` protocol has limited support for keyword-only
        argument. The only built-in means to pass kwargs is to the
        ``__setstate__`` method or through a callable with signature ``(obj,
        state)``. Neither of these methods allow for the kwargs to be passed to
        the constructor, only after the object is partially initialized. This
        does not work well for JAX, Equinox, or particularly run-time
        typechecking. To get around this, instead of the class type as the
        customary first element of reduced value we will instead return a
        `functools.partial`-wrapping of the class type with the kwargs bundled
        into the partial object. This is allowed by the ``__reduce__`` protocol,
        which says the first returned item must be any "callable object that
        will be called to create the initial version of the object." In
        conjunction with the standard args, the `functools.partial`-wrapped
        class type will properly construct any parametric subclass.

        Returns
        -------
        functools.partial[type]
            The `plum.type_unparametrized` form of this class object, which is
            the parametric class without the specialization to the specific
            parameter. The specialized subtype cannot be pickled since it is
            dynamically produced. The type is wrapped into a `functools.partial`
            with the keyword-only arguments to the constructor.
        tuple[Any, ...]
            The arguments to this class. Note: the keyword-only arguments are
            bundled with the type.

        Examples
        --------
        >>> import pickle
        >>> from unxt import Quantity
        >>> x = Quantity([1, 2, 3], "m")
        >>> pickle.dumps(x)
        b'...'

        """
        args, kwargs = [], {}
        for f in fields(self):
            v = getattr(self, f.name)
            if f.kw_only:
                kwargs[f.name] = v
            else:
                args.append(v)

        return partial(type_unparametrized(self), **kwargs), tuple(args)

    def __repr__(self) -> str:
        # --- class name ---
        base_cls_name = type_nonparametric(self).__name__
        # fmt: off
        if self._type_parameter == "unknown":
            try:  # Astropy v6-
                ptid = self._type_parameter._physical_type_id  # noqa: SLF001
            except AttributeError:  # Astropy v7+
                ptid = self._type_parameter._unit._physical_type_id  # noqa: SLF001
            dim = " ".join(
                f"{unit}{power}" if power != 1 else unit
                for unit, power in ptid
            )
        else:
            dim = self._type_parameter._name_string_as_ordered_set().split("'")[1]  # noqa: SLF001
        cls_name = f"{base_cls_name}[{dim!r}]"

        # --- args, kwargs ---

        fs = dict(field_items(self))
        del fs["value"]
        del fs["unit"]

        base_fields = f"{self.value!r}, unit={self.unit.to_string()!r}"
        extra_fields = ", ".join(f"{k}={v}" for k, v in fs.items())
        all_fields = base_fields + (", " + extra_fields if fs else "")

        # ------
        return f"{cls_name}({all_fields})"
        # fmt: on



"""Realizations of unit systems."""
# pylint: disable=no-member

__all__ = [
    # unit system instance
    "galactic",
    "dimensionless",
    "solarsystem",
    "si",
    # unit system alias
    "NAMED_UNIT_SYSTEMS",
]

import astropy.units as u

from .base import AbstractUnitSystem
from .builtin import (
    CGSUnitSystem,
    DimensionlessUnitSystem,
    LTMAUnitSystem,
    SIUnitSystem,
)

# Dimensionless. This is a singleton.
dimensionless = DimensionlessUnitSystem()

# Galactic unit system
galactic = LTMAUnitSystem(
    length=u.kpc,
    time=u.Myr,
    mass=u.Msun,
    angle=u.radian,
    # preferred_units={speed: u.km / u.s},
)

# Solar system units
solarsystem = LTMAUnitSystem(length=u.au, time=u.yr, mass=u.Msun, angle=u.radian)

# International System of Units default
si = SIUnitSystem(
    length=u.meter,
    time=u.second,
    mass=u.kilogram,
    electric_current=u.ampere,
    temperature=u.Kelvin,
    amount=u.mole,
    luminous_intensity=u.candela,
    angle=u.radian,
)

# Centimeter, gram, second
cgs = CGSUnitSystem(
    length=u.centimeter,
    time=u.second,
    mass=u.gram,
    angle=u.radian,
    force=u.dyne,
    energy=u.erg,
    pressure=u.barye,
    dynamic_viscosity=u.poise,
    kinematic_viscosity=u.stokes,
)


NAMED_UNIT_SYSTEMS: dict[str, AbstractUnitSystem] = {
    "galactic": galactic,
    "solarsystem": solarsystem,
    "dimensionless": dimensionless,
    "si": si,
    "cgs": cgs,
}



"""Orthogonal Mixin classes for Quantity classes."""

from collections.abc import Callable, Sequence
from typing import Any, cast

import numpy as np
from astropy.units import CompositeUnit
from jaxtyping import Array, ArrayLike

from dataclassish import replace

from .api import uconvert, ustrip
from unxt._src.units.core import AbstractUnits


class AstropyQuantityCompatMixin:
    """Mixin for compatibility with `astropy.units.Quantity`."""

    value: ArrayLike
    unit: AbstractUnits
    to_units: Callable[[Any], "AbstractQuantity"]
    to_units_value: Callable[[Any], ArrayLike]

    def to(self, u: Any, /) -> "AbstractQuantity":
        """Convert the quantity to the given units.

        See `AbstractQuantity.to_units`.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity(1, "m")
        >>> q.to("cm")
        Quantity['length'](Array(100., dtype=float32, ...), unit='cm')

        """
        return uconvert(u, self)  # redirect to the standard method

    def to_value(self, u: Any, /) -> ArrayLike:
        """Return the value in the given units.

        See `AbstractQuantity.to_units_value`.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity(1, "m")
        >>> q.to_value("cm")
        Array(100., dtype=float32, weak_type=True)

        """
        return ustrip(u, self)  # redirect to the standard method

    # TODO: support conversion of elements to Unit
    def decompose(self, bases: Sequence[AbstractUnits], /) -> "AbstractQuantity":
        """Decompose the quantity into the given bases.

        Examples
        --------
        >>> import astropy.units as u
        >>> from unxt import Quantity

        >>> q = Quantity(1, "m")
        >>> q.decompose([u.cm, u.s])
        Quantity['length'](Array(100., dtype=float32, ...), unit='cm')

        """
        du = self.unit.decompose(bases)  # decomposed units
        base_units = CompositeUnit(scale=1, bases=du.bases, powers=du.powers)
        return replace(self, value=self.value * du.scale, unit=base_units)


#####################################################################


SUPPORTED_IPYTHON_REPR_FORMATS: dict[str, str] = {
    "text/plain": "__repr__",
    "text/html": "_repr_html_",
    "text/latex": "_repr_latex_",
}


class IPythonReprMixin:
    """Mixin class for IPython representation of a Quantity."""

    value: Array
    unit: AbstractUnits

    def _repr_mimebundle_(
        self,
        *,
        include: Sequence[str] | None = None,
        exclude: Sequence[str] | None = None,
    ) -> dict[str, str]:
        r"""Return a MIME bundle representation of the Quantity.

        Parameters
        ----------
        include, exclude : Sequence[str] | None, optional
            The set of keys to include / exclude in the MIME bundle. If not
            provided, all supported formats are included. 'include' has
            precedence over 'exclude'.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity([1., 2, 3, 4], "m")
        >>> q._repr_mimebundle_()
        {'text/plain': "Quantity['length'](Array([1., 2., 3., 4.], dtype=float32), unit='m')",
         'text/html': '<span>[1., 2., 3., 4.]</span> * <span>Unit("m")</span>',
         'text/latex': '$[1.,~2.,~3.,~4.] \\; \\mathrm{m}$'}

        >>> q._repr_mimebundle_(include=["text/plain"])
        {'text/plain': "Quantity['length'](Array([1., 2., 3., 4.], dtype=float32), unit='m')"}

        >>> q._repr_mimebundle_(exclude=["text/html", "text/latex"])
        {'text/plain': "Quantity['length'](Array([1., 2., 3., 4.], dtype=float32), unit='m')"}

        """  # noqa: E501
        # Determine the set of keys to include in the MIME bundle
        keys: Sequence[str]
        if include is None and exclude is None:
            keys = tuple(SUPPORTED_IPYTHON_REPR_FORMATS)
        elif include is not None:
            keys = [key for key in include if key in SUPPORTED_IPYTHON_REPR_FORMATS]
        else:
            keys = [
                k for k in SUPPORTED_IPYTHON_REPR_FORMATS if k not in cast(str, exclude)
            ]

        # Create the MIME bundle
        return {
            key: getattr(self, SUPPORTED_IPYTHON_REPR_FORMATS[key])() for key in keys
        }

    def _repr_html_(self) -> str:
        """Return an HTML representation of the Quantity.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity([1., 2, 3, 4], "m")
        >>> q._repr_html_()
        '<span>[1., 2., 3., 4.]</span> * <span>Unit("m")</span>'

        """
        unit_repr = getattr(self.unit, "_repr_html_", self.unit.__repr__)()
        value_repr = np.array2string(self.value, separator=", ")

        return f"<span>{value_repr}</span> * <span>{unit_repr}</span>"

    def _repr_latex_(self) -> str:
        r"""Return a LaTeX representation of the Quantity.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity([1., 2, 3, 4], "m")
        >>> q._repr_latex_()
        '$[1.,~2.,~3.,~4.] \\; \\mathrm{m}$'

        """
        unit_repr = getattr(self.unit, "_repr_latex_", self.unit.__repr__)()
        value_repr = np.array2string(self.value, separator=",~")

        return f"${value_repr} \\; {unit_repr[1:-1]}$"

    # TODO: implement:
    # - _repr_markdown_
    # - _repr_json_



__all__ = [
    "AbstractUSysFlag",
    "StandardUSysFlag",
    "DynamicalSimUSysFlag",
]

from typing import Any


class AbstractUSysFlag:
    """Abstract class for unit system flags to provide dispatch control."""

    def __new__(cls, *_: Any, **__: Any) -> None:  # type: ignore[misc]
        msg = "unit system flag classes cannot be instantiated."
        raise ValueError(msg)


class StandardUSysFlag(AbstractUSysFlag):
    """Flag to indicate a standard unit system with no additional arguments."""


class DynamicalSimUSysFlag(AbstractUSysFlag):
    """Flag to indicate a unit system with optional definition of G."""
